#!/bin/sh

DBAG_ROOT="$(dirname "$(readlink -f "$0")")"
DBAG_PATH="${DBAG_ROOT}/src"
if [ -z "${PYTHONPATH}" ]; then
  PYTHONPATH="${DBAG_PATH}"
else
  PYTHONPATH="${DBAG_PATH}:${PYTHONPATH}"
fi

export PYTHONPATH
python3 -m deb_build_artifact_gather "$@"
