import dataclasses


@dataclasses.dataclass(slots=True, frozen=True)
class Tag:
    label: str


TAG_ICE = Tag("ICE")
TAG_GCC = Tag("gcc")
TAG_COMPILER = Tag("compiler")
TAG_AUTOCONF = Tag("autoconf")
TAG_CMAKE = Tag("cmake")
TAG_MESON = Tag("meson")
TAG_CONFIG_LOG = Tag("configure-log")
TAG_TEST_LOG = Tag("test-log")
