import dataclasses


@dataclasses.dataclass(frozen=True, slots=True)
class Variable:
    variable_name: str
    default_value: str | None = None


VARIABLE_BUILD_DIR = Variable(
    "BUILD_DIR",
    default_value=".",
)
